-- Constants
local ADDON_NAME = "GearGrinder"
local VALOR_CURRENCY_ID = 396
local JUSTICE_CURRENCY_ID = 395
local ELDER_CHARM_CURRENCY_ID = 697
local LESSER_CHARM_CURRENCY_ID = 738
local MOGU_CHARM_CURRENCY_ID = 752

local WHITE_BRIGHT_COLOR = "|cffFFFFFF"
local YELLOW_NEON_COLOR = "|cffFFFF00"
local RED_NEON_COLOR = "|cffFF0000"
local ORANGE_NEON_COLOR = "|cffEE6622"
local GREEN_NEON_COLOR = "|cff00FF00"
local GREY_COLOR = "|cff9d9d9d"
local END_COLOR = FONT_COLOR_CODE_CLOSE
local INFO_COLOR = "|cFFFFCC00"

-- list of dungeonIDs of interest to the addon
--/script for i=1, GetNumRFDungeons() do local a,b,_,_,_,_,_,_,_,c = GetRFDungeonInfo(i); print(i, a, b, c) end
local LFR_DUNGEONIDS = {
				MVa = 527,
				MVb = 528,
				HoFa = 529,
				HoFb = 530,
				ToES = 526,
				ToTa = 610,
				ToTb = 611,
				ToTc = 612,
				ToTd = 613};
				
local LFR_STARTER_DUNGEON = 527;
				
local LFR_CELL_A_DUNGEONS = {
				MVa = 527,
				MVb = 528};

local LFR_CELL_B_DUNGEONS = {
				HoFa = 529,
				HoFb = 530};

local LFR_CELL_C_DUNGEONS = {
				ToES = 526};
				
local LFR_CELL_D_DUNGEONS = {
				ToTa = 610,
				ToTb = 611,
				ToTc = 612,
				ToTd = 613};
			
local WORLD_BOSSES = {
				aSha = 32099,
				bGalleon = 32098,
				cNalak = 32518,
				dOondasta = 32519};
			
local CLASS_COLOR = {
	DEATHKNIGHT = "|cFFC41F3B",
	DRUID = "|cFFFF7D0A",
	HUNTER = "|cFFABD473",
	MAGE = "|cFF69CCF0",
	MONK = "|cFF00FF96",
	PALADIN = "|cFFF58CBA",
	PRIEST = "|cFFFFFFFF",
	ROGUE = "|cFFFFF569",
	SHAMAN = "|cFF0070DE",
	WARLOCK = "|cFF9482C9",
	WARRIOR = "|cFFC79C6E"
}

local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, false)


-- Start Titan VP Cap Progress
GG = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")
GG.version = GetAddOnMetadata(ADDON_NAME, "Version")
GG.category = GetAddOnMetadata(ADDON_NAME, "X-Category")
GG.versionMajor = 1;    -- Used for DB versioning

-- Load Libs
GG.LSM = LibStub:GetLibrary("LibSharedMedia-3.0")
GG.aceConfig = LibStub:GetLibrary("AceConfig-3.0")
GG.aceConfigDialog = LibStub:GetLibrary("AceConfigDialog-3.0")
GG.QTip = LibStub("LibQTip-1.0")
GG.LDB = LibStub("LibDataBroker-1.1")
GG.LDBIcon = LibStub("LibDBIcon-1.0")
GG.CBH = LibStub("CallbackHandler-1.0"):New(GG)

GG.currentPlayer = UnitName('player')	-- name of the logged-in character.
GG.currentRealm = GetRealmName()		-- current Realm we're on.
_, GG.currentClass = UnitClass("player") -- class of the current character.
GG.menu = false       -- Menu frame
GG.menuItems = {       -- Table for the dropdown menu
    top = {},
    middle = {},
    bottom = {},
}
GG.periodMaxVP = -1
GG.maxJP = -1
GG.maxVP = -1
GG.timezone = time({year=1970, month=1, day=2, hour=0})-86400

GG.LDBObj = {
    type = "data source",
		icon = "Interface\\Icons\\pvecurrency-valor",
		iconCoords = { 0.075, 0.925, 0.075, 0.925 },
    label = L.core.buttonLabel,
    text = 'n/a',
    category = GG.category,
    version = GG.version,
		OnEnter = function(frame)
  		-- if disabled during combat then don't show anything
      if InCombatLockdown() ~= nil and InCombatLockdown() == 1 and not GG:GetShowCombat() then return end
      GG:SetTooltipText(frame)
		end,
		OnLeave = function(frame) 
		end,
		OnClick = function(frame, button)
			if button == "RightButton" then
				GG:ShowOptions()
			end

      if button == "LeftButton" then
        if not GG:GetLDBDisplayLock() then GG:RotateButtonText() end
      end
		end
	}
	
-- OnLoad
function GG:OnInitialize()

    -- Make sure everything is ok with the db
    if not GG_Storage or type(GG_Storage) ~= 'table' then GG_Storage = {} end
    
    -- Tell the player we are being loaded
	self:Print(format(L.core.load, self.version))
        
    -- Load settings
    self.db = LibStub("AceDB-3.0"):New("GG_Settings", GG_Defaults, true)
	self:UpdateSettings()

	-- Set back to a launcher if text is off
	if not self:GetLDB() then
		self.LDBObj.type = 'launcher'
		self.LDBObj.text = ''
	end
        -- Start LDB
	self.LDBObj = self.LDB:NewDataObject(ADDON_NAME, self.LDBObj)

	-- Start the minimap icon
	self.LDBIcon:Register(ADDON_NAME, self.LDBObj, self.db.global.minimap)
    
    -- Register Options
	--GG_Options.args.purge.desc = format(L.core.options.purgeDesc, self:GetPurge() / 24)
	self.aceConfig:RegisterOptionsTable(L.core.name, GG_Options, {"GG", "GearGrinder"})
	self.aceConfigDialog:AddToBlizOptions(L.core.name)
	
    
    -- Events
	self:RegisterEvent("MERCHANT_CLOSED", function() GG:UpdateLDB(true) end)
	self:RegisterEvent("PLAYER_ENTERING_WORLD", function() GG:UpdateLDB(true) end)
	self:RegisterEvent("UPDATE_INSTANCE_INFO", function() GG:UpdateLDB(true) end)
	self:RegisterEvent("PLAYER_LEAVING_WORLD", function() GG:UpdateLDB(true) end)
	self:RegisterEvent("LFG_COMPLETION_REWARD", function() GG:UpdateLDB(true) end)
	self:RegisterEvent("PLAYER_REGEN_ENABLED", function() GG:UpdateLDB(true) end)
    self:RegisterEvent("CURRENCY_DISPLAY_UPDATE", function() GG:UpdateLDB(true) end)

    -- Add to Paperdoll
    table.insert(PAPERDOLL_STATCATEGORIES["GENERAL"].stats, L.core.PDVPText);
	table.insert(PAPERDOLL_STATCATEGORIES["GENERAL"].stats, L.core.PDVPCapText);
	table.insert(PAPERDOLL_STATCATEGORIES["GENERAL"].stats, L.core.PDJPText);
	
	if self:GetVPPaperdoll() then
		PAPERDOLL_STATINFO[L.core.PDVPText] = { updateFunc = function(...) GG:UpdateVPPaperDollFrame(...); end };
	else
		PAPERDOLL_STATINFO[L.core.PDVPText] = nil;
	end

	if self:GetVPCapPaperdoll() then
		PAPERDOLL_STATINFO[L.core.PDVPCapText] = { updateFunc = function(...) GG:UpdateVPCapPaperDollFrame(...); end };
	else
		PAPERDOLL_STATINFO[L.core.PDVPCapText] = nil;
	end

	if self:GetJPPaperdoll() then
		PAPERDOLL_STATINFO[L.core.PDJPText] = { updateFunc = function(...) GG:UpdateJPPaperDollFrame(...); end };
	else
		PAPERDOLL_STATINFO[L.core.PDJPText] = nil;
	end
	
    -- Build the menu
    --self:MenuInitialize();
    
    -- Get working on a score for the player
    self:UpdateLDB(); -- This may cause excesive loading time...
    
end

local bgFrame = {insets = {}}
function GG:UpdateBackdrop(tt)
	bgFrame.bgFile = self.LSM:Fetch("background", self.db.global.Background.Texture)
	bgFrame.edgeFile = self.LSM:Fetch("border", self.db.global.Background.BorderTexture)
	bgFrame.tile = self.db.global.Background.Tile
	bgFrame.tileSize = self.db.global.Background.TileSize
	bgFrame.edgeSize = self.db.global.Background.EdgeSize
	local inset = floor(self.db.global.Background.EdgeSize / 4)
	bgFrame.insets.left = inset
	bgFrame.insets.right = inset
	bgFrame.insets.top = inset
	bgFrame.insets.bottom = inset
	tt:SetBackdrop(bgFrame)

	local c = self.db.global.Background.Color
	tt:SetBackdropColor(c.r, c.g, c.b, c.a)

	c = self.db.global.Background.BorderColor
	tt:SetBackdropBorderColor(c.r, c.g, c.b, c.a)
	--tt:SetBackdropBorderColor(100, 100, 100, 100)
end

function GG:SetTooltipText(f)
    -- Tooltip title
	local val = format(L.core.menuText, format(L.core.versionText, self.version));
	local num_columns, line = 6, 0;
	
	if self.LDBtip then self.QTip:Release(self.LDBtip) end
	self.LDBtip = self.QTip:Acquire(ADDON_NAME .. "TT", num_columns, "LEFT", "RIGHT", "RIGHT", "RIGHT", "RIGHT", "RIGHT")
	tt = self.LDBtip
	tt.anchorframe = f
	tt:EnableMouse()
	tt:SetScale(self:GetTTScale())
	tt:SetAutoHideDelay(0.1, f)
	tt:SmartAnchorTo(f)
	--tt:SetAutoHideDelay(0.1, f)	
	tt:Clear()
	
	GG:UpdateBackdrop(tt)
	
	line = tt:AddHeader(".")
	tt:SetCell(line, 1, val, "CENTER", num_columns)
	--tt:SetText(val, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);

	-- calculate the time remaining.
	local sec = time()
	local dtDate = time({year=date('%Y',sec),month=date('%m',sec),day=date('%d',sec),hour=0})

	local dtPrev = dtDate - (((date("%w", dtDate) - self:GetResetDayOfWeek()) % 7) ) * 86400 + (self:GetResetHour() * 60 * 60)
	if dtPrev > sec then
	  dtPrev = dtPrev - (7 * 24 * 60 * 60)
	end
	local dtNext = dtPrev + 7 * 24 * 60 * 60
 	local timeRemaining = dtNext - sec

	line = tt:AddSeparator()
	
	line = tt:AddLine("" .. L.core.cooldownInfoText, ".")
	val = WHITE_BRIGHT_COLOR .. self:GetRemainingTimeText(timeRemaining) .. END_COLOR
	tt:SetCell(line, 2, val, "LEFT", num_columns-1)
	
	if self:GetDebug() then
		line = tt:AddLine("Now: " .. WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", time()) .. "   " .. time() .. END_COLOR)
		--tt:AddLine("  Last Reset: " .. WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", dtPrev) .. "   " .. dtPrev .. END_COLOR);
		--tt:AddLine("  Date: " .. WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", dtDate) .. "   " .. dtDate .. END_COLOR);
		--tt:AddLine("  Time: " .. WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", sec) .. "   " .. sec .. END_COLOR);
		--tt:AddLine("  Resets: " .. WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", dtNext) .. "   " .. dtNext .. END_COLOR);
	end
	line = tt:AddLine("" .. L.core.cooldownInfoLastText, ".")
	val = WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", dtPrev) .. END_COLOR
	tt:SetCell(line, 2, val, "LEFT", num_columns-1)
	
	line = tt:AddLine("" .. L.core.cooldownInfoNextText, ".");
	val = WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", dtNext) .. END_COLOR
	tt:SetCell(line, 2, val, "LEFT", num_columns-1)
	
	--tt:AddLine("  Date: " .. WHITE_BRIGHT_COLOR  .. date("%d-%b-%Y %I:%M %p", dtDate) .. "   " .. dtDate .. END_COLOR);
	line = tt:AddLine(" ")
	--tt:AddSeparator();
	
	if self:GetWorldBoss() then
		self:LoadAltBoss(tt)
		tt:AddLine(" ")
	end
	
	if self:GetLFR() then
		self:LoadAltLFR(tt)
		tt:AddLine(" ")
	end
	
	self:LoadAlt(tt)

	-- add the info lines at the end
	-- add the "Toggle the display of <toon>" line at the end
	
	if self:GetFiltering() then
		line = tt:AddLine(" ")
		line = tt:AddLine(".", ".")
		tt:SetCell(line, 1, INFO_COLOR .. "Click to toggle the display of " .. CLASS_COLOR[self.currentClass] .. self.currentPlayer .. END_COLOR, "CENTER", num_columns)
		tt:SetCellScript(line, 1, "OnMouseUp", function(frame) GG:ToggleToonFilter(); GG.LDBtip:Hide() end)
	end
	
	tt:Show()
end

function GG:LoadAltBoss(tt)
	local sortorder;

	line = tt:AddHeader(L.core.BossTitle, "Sha", "Galleon", "Nalak", "Oondasta");	
	tt:SetCell(line, 2, L.core.BossShaOfAngerShort, "CENTER")
	tt:SetCell(line, 3, L.core.BossGalleonShort, "CENTER")
	tt:SetCell(line, 4, L.core.BossNalakShort, "CENTER")
	tt:SetCell(line, 5, L.core.BossOondastaShort, "CENTER")
	
	sortorder = self:GetSortOrder();
	
	if sortorder == 1 then  -- Unsorted
	
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do
			self:ShowAltBoss(toonname, alt, tt)
		end
	
	elseif sortorder == 2 then  -- Alphabetical
		for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
			self:ShowAltBoss(toonname, alt, tt)
		end
	
	elseif sortorder == 3 then  -- Class
		for class in orderedPairs(CLASS_COLOR) do
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.class == class then
					self:ShowAltBoss(toonname, alt, tt)
				end
			end
		end
	
	elseif sortorder == 4 then  -- Valor Points
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyVP) then
				table.insert(tmp, alt.qtyVP)
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(G_Storage[self.currentRealm]) do
				if alt.qtyVP == v then
					self:ShowAltBoss(toonname, alt, tt);
				end
			end
		end
		
	elseif sortorder == 5 then  -- iLevel
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.iLvl) then
				table.insert(tmp, alt.iLvl);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.iLvl == v then
					self:ShowAltBoss(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 6 then  -- Justice Points
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyJP) then
				table.insert(tmp, alt.qtyJP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyJP == v then
					self:ShowAltBoss(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 7 then  -- Charms
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyLCharm) then
				table.insert(tmp, alt.qtyLCharm);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyLCharm == v then
					self:ShowAltBoss(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 8 then  -- VP Cap Progress
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.periodEarnVP) then
				table.insert(tmp, alt.periodEarnVP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.periodEarnVP == v then
					self:ShowAltBoss(toonname, alt, tt);
				end
			end
		end

	end
	
end

function GG:ShowAltBoss(toon, alt, tt)
	local toonColor, toonText = "", ""
	local line, cell = 0, 0
	local sec = time()
	local dtDate = time({year=date('%Y',sec),month=date('%m',sec),day=date('%d',sec),hour=0})
	local boss = {}
	local showtoon
	local dtPrev = dtDate - (((date("%w", dtDate) - self:GetResetDayOfWeek()) % 7) ) * 86400 + (self:GetResetHour() * 60 * 60)
	
	if dtPrev > sec then
	  dtPrev = dtPrev - (7 * 24 * 60 * 60)
	end
	
	if alt.active == nil then alt.active = true end
	if self:GetFiltering() and not alt.active then return end  -- exit function if the toon is not active
	
	if alt.qtyJP == nil then alt.qtyJP = 0 end
	if alt.qtyVP == nil then alt.qtyVP = 0 end
	
	-- by default the toon will not show
	showtoon = false

	-- one time check to setup the database for the toon.
	if alt.boss == nil then
		alt.boss = {}
		for i, v in pairs(WORLD_BOSSES) do
			alt.boss[v] = false
		end
		alt.boss1, alt.boss2, alt.boss3, alt.boss4 = nil,nil,nil,nil
	end
	
	-- setup the boss values
	if dtPrev < alt.savingTime then
		for i, v in pairs(WORLD_BOSSES) do
			boss[v] = alt.boss[v]
			-- if a boss has been killed then show the line
			if boss[v] then showtoon = true end
		end
	end

	-- if the toon can't do the first raid then ignore
	if not showtoon 
	   and alt.dungeon ~= nil 
	   and alt.dungeon[LFR_STARTER_DUNGEON] ~= nil 
	   and alt.dungeon[LFR_STARTER_DUNGEON].isAvailable ~= nil
	   and alt.dungeon[LFR_STARTER_DUNGEON].isAvailable == 1 then showtoon = true end
	
	-- if there is a value for any of the currencies then show the line
	if showtoon or self:GetDebug() then	
		
		-- determine the class color
		toonColor = CLASS_COLOR[alt.class]
		toonText = toonColor .. " " .. toon .. END_COLOR;
		
		line = tt:AddLine(toonText, ".")
		cell = 2
		for i, v in orderedPairs(WORLD_BOSSES) do
			if boss[v] then 
				bossText = RED_NEON_COLOR .. L.core.BossLooted .. END_COLOR
				if self:GetShowIcon() then bossText = bossText .. " |TInterface\\Icons\\Achievement_pvp_h_06:16:16|t" end
			else 
				bossText = GREEN_NEON_COLOR .. L.core.BossNotLooted .. END_COLOR 
				if self:GetShowIcon() then bossText = bossText .. " |TInterface\\Icons\\Achievement_pvp_g_06:16:16|t" end
			end
			tt:SetCell(line, cell, bossText, "CENTER");
			cell = cell + 1
		end
	end
	
end

function GG:LoadAltLFR(tt)
	local sortorder;

	line = tt:AddHeader(L.core.LFRTitle, ".", ".", ".", ".");	
	tt:SetCell(line, 2, L.core.LFR_MV, "CENTER")
	tt:SetCell(line, 3, L.core.LFR_HoF, "CENTER")
	tt:SetCell(line, 4, L.core.LFR_ToES, "CENTER")
	tt:SetCell(line, 5, L.core.LFR_ToT, "CENTER")
	
	sortorder = self:GetSortOrder();
	
	if sortorder == 1 then  -- Unsorted
	
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do
			self:ShowAltLFR(toonname, alt, tt);
		end
	
	elseif sortorder == 2 then  -- Alphabetical
		for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
			self:ShowAltLFR(toonname, alt, tt);
		end
	
	elseif sortorder == 3 then  -- Class
		for class in orderedPairs(CLASS_COLOR) do
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.class == class then
					self:ShowAltLFR(toonname, alt, tt);
				end
			end
		end
	
	elseif sortorder == 4 then  -- Valor Points
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyVP) then
				table.insert(tmp, alt.qtyVP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyVP == v then
					self:ShowAltLFR(toonname, alt, tt);
				end
			end
		end
		
	elseif sortorder == 5 then  -- iLevel
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.iLvl) then
				table.insert(tmp, alt.iLvl);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.iLvl == v then
					self:ShowAltLFR(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 6 then  -- Justice Points
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyJP) then
				table.insert(tmp, alt.qtyJP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyJP == v then
					self:ShowAltLFR(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 7 then  -- Charms
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyLCharm) then
				table.insert(tmp, alt.qtyLCharm);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyLCharm == v then
					self:ShowAltLFR(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 8 then  -- VP Cap Progress
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.periodEarnVP) then
				table.insert(tmp, alt.periodEarnVP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.periodEarnVP == v then
					self:ShowAltLFR(toonname, alt, tt);
				end
			end
		end

	end
	
end

function GG:ShowAltLFR(toon, alt, tt)
	local toonColor, toonText = "", "";
	local cellText = "";
	
	local dungeons = {};
	
	local sec = time();
	local dtDate = time({year=date('%Y',sec),month=date('%m',sec),day=date('%d',sec),hour=0});
	
	local dtPrev = dtDate - (((date("%w", dtDate) - self:GetResetDayOfWeek()) % 7) ) * 86400 + (self:GetResetHour() * 60 * 60)
	if dtPrev > sec then
	  dtPrev = dtPrev - (7 * 24 * 60 * 60);
	end
	
	if alt.active == nil then alt.active = true end
	if self:GetFiltering() and not alt.active then return end  -- exit function if the toon is not active

	-- if the data has not been saved then ignore
	if alt.dungeon == nil then return end
	
	-- if the toon can't do the first raid then ignore
	if alt.dungeon[LFR_STARTER_DUNGEON] == nil then return end
	if alt.dungeon[LFR_STARTER_DUNGEON].isAvailable == nil then return end
	if alt.dungeon[LFR_STARTER_DUNGEON].isAvailable <= 0 then return end
	
	-- determine the class color
	toonColor = CLASS_COLOR[alt.class]
	toonText = toonColor .. " " .. toon .. END_COLOR;
	
	if dtPrev < alt.savingTime then
		-- use the data in the database
		dungeons = alt.dungeon
	else
		--ignore the data in the database
		for i, v in pairs(alt.dungeon) do
			dungeons[i] = {}
			dungeons[i].isComplete = 0
			dungeons[i].qtyLooted = 0
			dungeons[i].isAvailable = v.isAvailable
		end
	end

	line = tt:AddLine(toonText, ".");

	GG:SetLFRCell(tt, LFR_CELL_A_DUNGEONS, dungeons, line, 2)
	GG:SetLFRCell(tt, LFR_CELL_B_DUNGEONS, dungeons, line, 3)
	GG:SetLFRCell(tt, LFR_CELL_C_DUNGEONS, dungeons, line, 4)
	GG:SetLFRCell(tt, LFR_CELL_D_DUNGEONS, dungeons, line, 5)
	
end

function GG:LoadAlt(tt)

	local sortorder;

	line = tt:AddHeader(L.core.PDCharTextShort);
	if self:GetiLevel() then
		tt:SetCell(line, 2, L.core.PDiLvlTextShort, "CENTER")
	end
	tt:SetCell(line, 3, L.core.PDVPCapTextShort, "CENTER")
	tt:SetCell(line, 4, L.core.PDVPTextShort, "CENTER")
	tt:SetCell(line, 5, L.core.PDJPTextShort, "CENTER")
	if self:GetCharms() then 
		tt:SetCell(line, 6, L.core.PDCharmTextShort, "CENTER")
	end
		
	sortorder = self:GetSortOrder();
	
	if sortorder == 1 then  -- Unsorted
	
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do
			self:ShowAlt(toonname, alt, tt);
		end
	
	elseif sortorder == 2 then  -- Alphabetical
		for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
			self:ShowAlt(toonname, alt, tt);
		end
	
	elseif sortorder == 3 then  -- Class
		for class in orderedPairs(CLASS_COLOR) do
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.class == class then
					self:ShowAlt(toonname, alt, tt);
				end
			end
		end
	
	elseif sortorder == 4 then  -- Valor Points
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyVP) then
				table.insert(tmp, alt.qtyVP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyVP == v then
					self:ShowAlt(toonname, alt, tt);
				end
			end
		end
		
		
	elseif sortorder == 5 then  -- iLevel
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.iLvl) then
				table.insert(tmp, alt.iLvl);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.iLvl == v then
					self:ShowAlt(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 6 then  -- Justice Points
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyJP) then
				table.insert(tmp, alt.qtyJP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyJP == v then
					self:ShowAlt(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 7 then  -- Charms
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.qtyLCharm) then
				table.insert(tmp, alt.qtyLCharm);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.qtyLCharm == v then
					self:ShowAlt(toonname, alt, tt);
				end
			end
		end

	elseif sortorder == 8 then  -- VP Cap Progress
		tmp = {};
		for toonname, alt in pairs(GG_Storage[self.currentRealm]) do 
			if not table.contains(tmp, alt.periodEarnVP) then
				table.insert(tmp, alt.periodEarnVP);
			end
		end
		table.sort(tmp, function(a,b) return a>b end);	
		for i, v in ipairs(tmp) do 
			for toonname, alt in orderedPairs(GG_Storage[self.currentRealm]) do
				if alt.periodEarnVP == v then
					self:ShowAlt(toonname, alt, tt);
				end
			end
		end

	end
	
end

function GG:ShowAlt(toon, alt, tt)
	local toonColor, toonText, valCapColor, valVPColor, valJPColor, valECharmColor, valCapText, valCharmText, valCharmColor = "", "", "", "", "", "", "", "", "";
	local valVP, valJP, valCap, valiLvl, valiELvl, valECharm, valLCharm, valMCharm = 0,0,0,0,0,0,0,0;

	local sec = time();
	local dtDate = time({year=date('%Y',sec),month=date('%m',sec),day=date('%d',sec),hour=0});

	local dtPrev = dtDate - (((date("%w", dtDate) - self:GetResetDayOfWeek()) % 7) ) * 86400 + (self:GetResetHour() * 60 * 60)
	if dtPrev > sec then
	  dtPrev = dtPrev - (7 * 24 * 60 * 60);
	end

	if alt.active == nil then alt.active = true end
	if self:GetFiltering() and not alt.active then return end  -- exit function if the toon is not active
	
	-- if there is a value for any of the currencies then show the line
	if alt.qtyVP > 0 or alt.qtyJP > 0 or alt.qtyECharm > 0 or alt.qtyLCharm > 0 or alt.qtyMCharm > 0 or self:GetDebug() then

		if alt.qtyMaxVP == nil then
			alt.qtyMaxVP = self.db.global.VPStandardMax;
		elseif alt.qtyMaxVP <= 0 then
			alt.qtyMaxVP = self.db.global.VPStandardMax;
		end 

		if alt.periodMaxVP == nil then
			alt.periodMaxVP = self.db.global.VPStandardCapMax;
		elseif alt.periodMaxVP <= 0 then
			alt.periodMaxVP = self.db.global.VPStandardCapMax;
		end 

		if alt.qtyMaxJP == nil then
			alt.qtyMaxJP = self.db.global.JPStandardMax;
		elseif alt.qtyMaxJP <= 0 then
			alt.qtyMaxJP = self.db.global.JPStandardMax;
		end 
		
		if alt.iLvl == nil then alt.iLvl = 0 end
		if alt.iELvl == nil then alt.iELvl = 0 end		
		if alt.qtyECharm == nil then alt.qtyECharm = 0 end
		if alt.qtyLCharm == nil then alt.qtyLCharm = 0 end
		if alt.qtyMCharm == nil then alt.qtyMCharm = 0 end
	
		self:Debug("VP Max Cap: " .. alt.qtyMaxVP);
	
		-- determine the class color
		toonColor = CLASS_COLOR[alt.class]
		--local off_left, off_right, off_top, off_bottom = unpack(CLASS_ICON_TCOORDS[alt.class]);
		--local x = off_left .. ":" .. off_top .. ":" .. off_right - off_left+1 .. ":" .. off_bottom-off_top+1;
		--toonText = "|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:16:16:" .. x .. "|t" .. toonColor .. toon .. END_COLOR;
		toonText = toonColor .. toon .. END_COLOR;
		
		if dtPrev < alt.savingTime then valCap = alt.periodEarnVP else valCap = 0 end

		valiLvl = alt.iLvl
		valiELvl = alt.iELvl
		valVP = alt.qtyVP
		valJP = alt.qtyJP
		valECharm = alt.qtyECharm
		valLCharm = alt.qtyLCharm
		valMCharm = alt.qtyMCharm

		valCapColor = self:GetProgressColor(valCap, alt.periodMaxVP);
		valVPColor = self:GetProgressColor(valVP, alt.qtyMaxVP);
		valJPColor = self:GetProgressColor(valJP, alt.qtyMaxJP);
		valCharmColor = GREEN_NEON_COLOR;
		if self:GetiLevel() then
			valiLvlText = valiLvl 
			if self:GetShowIcon() then valiLvlText = valiLvlText .. " |TInterface\\Icons\\Inv_helmet_03:16:16|t" end
		else
			valiLvlText = nil
		end
		valCapText = valCapColor .. valCap .. END_COLOR -- .. "/" .. alt.periodMaxVP .. END_COLOR
		if self:GetShowIcon() then valCapText = valCapText .. " |TInterface\\Icons\\pvecurrency-valor:16:16|t" end
		valVPText = valVPColor .. valVP .. END_COLOR 
		if self:GetShowIcon() then valVPText = valVPText .. " |TInterface\\Icons\\pvecurrency-valor:16:16|t" end
		valJPText = valJPColor .. valJP .. END_COLOR 
		if self:GetShowIcon() then valJPText = valJPText .. " |TInterface\\Icons\\pvecurrency-justice:16:16|t" end
		if self:GetCharms() then
			valCharmText = valCharmColor .. valLCharm .. "/" .. valECharm .. "/" .. valMCharm .. END_COLOR 
			if self:GetShowIcon() then valCharmText = valCharmText .. " |TInterface\\Icons\\archaeology_5_0_mogucoin:16:16|t" end
		else
			valCharmText = nil
		end

		if self:GetDebug() then
			tt:AddLine(" " .. toonText,  valVPColor .. 1234 .. END_COLOR, valCapColor .. "1234/1234" .. END_COLOR);
		else
			tt:AddLine(" " .. toonText, valiLvlText, valCapText , valVPText, valJPText, valCharmText);
		end
	end
end

function GG:SetLFRCell(tt, dungeonList, data, line, cellID)

	
	cellText = ""
	for i, v in orderedPairs(dungeonList) do
		if cellText ~= "" then cellText = cellText .. " " end
		if data[v] ~= nil and data[v].isAvailable ~= nil then
			if data[v].isAvailable == 1 then
				if data[v].isComplete == 1 then
					-- red text
					cellText = cellText .. RED_NEON_COLOR
				else
					-- green text
					cellText = cellText .. GREEN_NEON_COLOR
				end
			else
				-- orange text
				cellText = cellText .. GREY_COLOR
			end
			cellText = cellText .. data[v].qtyLooted .. END_COLOR
		end
	end
	tt:SetCell(line, cellID, cellText, "CENTER");

end

function GG:GetProgressColor(val, maxval)
	if val >= maxval then 
		return RED_NEON_COLOR;
	elseif val > (maxval - self.db.global.progress_orange) then 
		return ORANGE_NEON_COLOR;
	elseif val > (maxval - self.db.global.progress_yellow) then 
		return YELLOW_NEON_COLOR;
	else 
		return GREEN_NEON_COLOR;
	end
	return GREEN_NEON_COLOR;
end

-- Save the data to the database.
function GG:SaveData()

	local qtyVP, qtyMaxVP, wk_earnVP, wk_MaxVP
	local qtyJP, wk_earnJP, qtyMaxJP
	local qtyECharm, qtyLCharm, qtyMCharm
	local iLvl, iELvl
	local a, b, c, d, e, f
	local data
	
	self:Debug("Saving Gear Grinder...");
	self:Debug("Current Player " .. self.currentPlayer);
	self:Debug("Current Realm " .. self.currentRealm);
	
	if GG_Storage == nil then
		GG_Storage = {};
	end
	if GG_Storage[self.currentRealm] == nil then
		GG_Storage[self.currentRealm] = {};
	end	
	if GG_Storage[self.currentRealm][self.currentPlayer] == nil then
		GG_Storage[self.currentRealm][self.currentPlayer] = {};
		data = GG_Storage[self.currentRealm][self.currentPlayer];
		local playerClass, englishClass = UnitClass("player");
		data.class = englishClass;
	else 
		data = GG_Storage[self.currentRealm][self.currentPlayer];
	end
	
	GG_Storage[self.currentRealm]["__orderedIndex"] = nil
	
	data.savingTime = time();
		
	_, qtyVP, _, wk_earnVP, wk_MaxVP, qtyMaxVP = GetCurrencyInfo(VALOR_CURRENCY_ID);
	_, qtyJP, _, wk_earnJP, _, qtyMaxJP = GetCurrencyInfo(JUSTICE_CURRENCY_ID);
	_, qtyECharm = GetCurrencyInfo(ELDER_CHARM_CURRENCY_ID);
	_, qtyLCharm = GetCurrencyInfo(LESSER_CHARM_CURRENCY_ID);
	_, qtyMCharm = GetCurrencyInfo(MOGU_CHARM_CURRENCY_ID);
	iLvl, iELvl = GetAverageItemLevel();
	
	wk_MaxVP = floor(wk_MaxVP/100);
	qtyMaxVP = floor(qtyMaxVP/100);
	qtyMaxJP = floor(qtyMaxJP/100);
	iLvl = floor(iLvl);
	iELvl = floor(iELvl);
	
	-- save data 
	-- New values that are recorded.
	data.qtyVP = qtyVP;
	data.qtyMaxVP = qtyMaxVP;
	data.periodEarnVP = wk_earnVP;
	data.periodMaxVP = wk_MaxVP;
	
	data.qtyJP = qtyJP;
	data.periodEarnJP = wk_earnJP; 
	data.qtyMaxJP = qtyMaxJP;
	
	data.qtyECharm = qtyECharm;
	data.qtyLCharm = qtyLCharm;
	data.qtyMCharm = qtyMCharm;
	
	if iLvl ~= 0 then
		data.iLvl = iLvl;
	end
	
	if iELvl ~= 0 then
		data.iELvl = iELvl;
	end

	-- save off the LFR Raid saves/completions
				
	for i, v in pairs(LFR_DUNGEONIDS) do
		_, numKilled = GetLFGDungeonNumEncounters(v);
		_, _, raidCompleted, raidAvailable = GetLFGDungeonRewardCapInfo(v);
	
		if data.dungeon == nil then data.dungeon = {} end
		if data.dungeon[v] == nil then data.dungeon[v] = {} end
		data.dungeon[v].qtyLooted = numKilled
		data.dungeon[v].isComplete = raidCompleted
		data.dungeon[v].isAvailable = raidAvailable
	end

	-- reset the values to future proof the addon
	data.boss1 = nil;
	data.boss2 = nil;
	data.boss3 = nil;
	data.boss4 = nil;
	
	-- save off the world boss kills
	if data.boss == nil then data.boss = {} end
	for i, v in pairs(WORLD_BOSSES) do
		local t={v}
		t = GetQuestsCompleted(t)
		
		data.boss[v] = t[v]
		if data.boss[v] == nil then data.boss[v] = false end
	end
	
	-- setup the filtered value
	if data.active == nil then data.active = true end
	
end


-- Make sure the database is the latest version
function GG:UpdateSettings()
    
    if self.db.global.version == self.versionMajor then
        -- Same version
    end
    
    -- Update version information
    self.db.global.version = self.versionMajor;
end

-- Purge the Database
function GG:PurgeDB()

	GG_Storage = nil;
	
	self:SaveData();

end

--[[

    Setters, Getter and Togglers

]]

-- Set
function GG:SetShowCombat(v) self.db.global.combat = v; end
function GG:SetCharms(v) self.db.global.charms = v; end
function GG:SetiLevel(v) self.db.global.iLevel = v; end
function GG:SetSortOrder(v) self.db.global.sortOrder = v; end
function GG:SetWorldBoss(v) self.db.global.worldBoss = v; end
function GG:SetShowIcon(v) self.db.global.show_Icons = v; end
function GG:SetLFR(v) self.db.global.lfr = v; end
function GG:SetTTScale(v) self.db.global.ttScale = v; end
function GG:SetFiltering(v) self.db.global.filtering = v; end
function GG:SetBGGeneric(i, v) self.db.global.Background[ i[#i] ] = v end
function GG:SetLabel(v) self.db.global.showLabel = v; self:UpdateLDB(true); end
function GG:SetLDBlabel(v) self.db.global.ldbLabel = v; self:UpdateLDB(true); end
function GG:SetLDBTextColor(v) self.db.global.ldbTextColor = v; self:UpdateLDB(true); end
function GG:SetLDBDisplay(v) self.db.global.ldbDisplay = v; self:UpdateLDB(true); end
function GG:SetLDBDisplayLock(v) self.db.global.ldbDisplayLock = v; end

-- Get
function GG:GetMinimap() return not self.db.global.minimap.hide; end
function GG:GetVPPaperdoll() return self.db.global.cinfoVP; end
function GG:GetVPCapPaperdoll() return self.db.global.cinfoVPCap; end
function GG:GetJPPaperdoll() return self.db.global.cinfoJP; end
function GG:GetLabel() return self.db.global.showLabel; end
function GG:GetLDB() return self.db.global.ldbText; end
function GG:GetLDBTextColor() return self.db.global.ldbTextColor; end
function GG:GetLDBlabel() return self.db.global.ldbLabel; end
function GG:GetDebug() return self.db.char.debug; end
function GG:GetCharms() return self.db.global.charms; end
function GG:GetShowCombat() return self.db.global.combat; end
function GG:GetiLevel() return self.db.global.iLevel; end
function GG:GetSortOrder() return self.db.global.sortOrder; end
function GG:GetWorldBoss() return self.db.global.worldBoss; end
function GG:GetShowIcon() return self.db.global.show_Icons; end
function GG:GetBGGeneric(i) return self.db.global.Background[ i[#i] ] end
function GG:GetBGTile() return self.db.global.Background.Tile end
function GG:GetLFR() return self.db.global.lfr end
function GG:GetTTScale() return self.db.global.ttScale end
function GG:GetFiltering() return self.db.global.filtering end
function GG:GetLDBDisplay() return self.db.global.ldbDisplay end
function GG:GetLDBDisplayLock() return self.db.global.ldbDisplayLock end

-- Toggle
function GG:ToggleShowIcon() self:SetShowIcon(not self:GetShowIcon()); end
function GG:ToggleMinimap() self:SetMinimap(not self:GetMinimap()); end
function GG:ToggleCharms() self:SetCharms(not self:GetCharms()); end
function GG:ToggleShowCombat() self:SetShowCombat(not self:GetShowCombat()); end
function GG:ToggleiLevel() self:SetiLevel(not self:GetiLevel()); end
function GG:ToggleWorldBoss() self:SetWorldBoss(not self:GetWorldBoss()); end
function GG:ToggleSortOrder() self:SetSortOrder(not self:GetSortOrder()); end
function GG:ToggleVPPaperdoll() self:SetVPPaperdoll(not self:GetVPPaperdoll()); end
function GG:ToggleVPCapPaperdoll() self:SetVPCapPaperdoll(not self:GetVPCapPaperdoll()); end
function GG:ToggleJPPaperdoll() self:SetJPPaperdoll(not self:GetJPPaperdoll()); end
function GG:ToggleLabel() self:SetLabel(not self:GetLabel()); end
function GG:ToggleLDBlabel() self:SetLDBlabel(not self:GetLDBlabel()); end
function GG:ToggleLDBTextColor() self:SetLDBTextColor(not self:GetLDBTextColor()); end
function GG:ToggleLFR() self:SetLFR(not self:GetLFR()); end
function GG:ToggleFiltering() self:SetFiltering(not self:GetFiltering()); end
function GG:ToggleLDBDisplayLock() self:SetLDBDisplayLock(not self:GetLDBDisplayLock()); end
function GG:ToggleToonFilter() self:SetToonFilter(not self:GetToonFilter()); end

-- Advanced Gets
function GG:GetBGColor(i)
	local t = self.db.global.Background.Color
	return t.r, t.g, t.b, t.a
end

function GG:GetBGBorderColor(i)
	local x = self.db.global.Background.BorderColor
	return x.r, x.g, x.b, x.a
end

function GG:GetRemainingTimeText(seconds)
	local days = floor(seconds / 86400);
	seconds = seconds - days * 86400;
	local hours = floor(seconds / 3600);
	seconds = seconds - hours * 3600;
	local minutes = floor(seconds / 60);
	seconds = seconds - minutes * 60;
	local TIME_COLOR = "";
	local TIME_TEXT = "";
	local timeStr = "";		-- one space just to pad slightly vs being too close to raid/instance name string.
	
	-- compact view's colors.
	-- moved "Reset in" text so that it's now above lockout time.
	TIME_COLOR = WHITE_BRIGHT_COLOR ;
	TIME_TEXT = WHITE_BRIGHT_COLOR ;
	
	if days > 0 then
		timeStr = timeStr .. TIME_COLOR .. days .. TIME_TEXT .. " " .. L.core.cooldownInfoDay .. END_COLOR .. " ";
	end
	
	if hours > 0 then
		timeStr = timeStr .. TIME_COLOR .. hours .. TIME_TEXT .. " " .. L.core.cooldownInfoHr .. END_COLOR .. " ";
	end

	if minutes then 
		timeStr = timeStr .. TIME_COLOR .. minutes .. TIME_TEXT .. " " .. L.core.cooldownInfoMin .. END_COLOR .. " ";
	end


	return timeStr;
end

function GG:RotateButtonText()

  -- increase the display value (rotate it to min value if at max)
  x = self:GetLDBDisplay()

  if x == 10 then x = 1 else x = x + 1 end

  self:SetLDBDisplay(x)

end

function GG:GetButtonText()
	local a, displaytype = "", self:GetLDBDisplay()
	local vpcapText, vpText, lfrText, jpText, finalText = "", "", "", "", ""
	local lfrDone, lfrTotal = 0, 0
	local data = GG_Storage[self.currentRealm][self.currentPlayer]
	local earnVP, qtyVP, qtyJP = data.periodEarnVP, data.qtyVP, data.qtyJP
	
--[[
						[1] = "VP Cap (VP) <JP> [LFR]",
						[2] = "VP Cap (VP) [LFR]",
						[3] = "VP Cap (VP) <JP>",
						[4] = "VP Cap (VP)",
						[5] = "VP Cap [LFR]", 
						[6] = "VP Cap",
						[7] = "VP <JP> [LFR]",
						[8] = "VP <JP>",
						[9] = "VP [LFR]",
						[10] = "LFR",
]]
						
	-- Determine the VPCap text to display
	if displaytype >= 1 and displaytype <= 6 then
		if self:GetPeriodMaxVP() <= 0 then
			vpcapText = "N/A";
		else
			vpcapText = earnVP .. "/" .. self:GetPeriodMaxVP()
		end

		if self:GetLDBTextColor() then
			if earnVP >= self:GetPeriodMaxVP() then 
				a = RED_NEON_COLOR
			elseif earnVP > (self:GetPeriodMaxVP() - self.db.global.progress_orange) then 
				a = ORANGE_NEON_COLOR
			elseif earnVP > (self:GetPeriodMaxVP() - self.db.global.progress_yellow) then 
				a = YELLOW_NEON_COLOR
			else
				a = GREEN_NEON_COLOR
			end
			vpcapText = a .. vpcapText .. END_COLOR
		end
	end
	
	-- Determine the VP text to display
	if (displaytype >= 1 and displaytype <= 4) or (displaytype >= 7 and displaytype <= 9) then
		if self:GetMaxVP() <= 0 then
			vpText = "N/A";
		else
			vpText = qtyVP
		end

		if self:GetLDBTextColor() then
			if qtyVP >= self:GetMaxVP() then 
				a = RED_NEON_COLOR
			elseif qtyVP > (self:GetMaxVP() - self.db.global.progress_orange) then 
				a = ORANGE_NEON_COLOR
			elseif qtyVP > (self:GetMaxVP() - self.db.global.progress_yellow) then 
				a = YELLOW_NEON_COLOR
			else
				a = GREEN_NEON_COLOR
			end
			vpText = a .. vpText .. END_COLOR
		end
	end
	
	-- Determine the JP text to display
	if displaytype == 1 or displaytype == 3 or displaytype == 7 or displaytype == 8 then
		if self:GetMaxJP() <= 0 then
			jpText = "N/A";
		else
			jpText = qtyJP
		end

		if self:GetLDBTextColor() then
			if qtyJP >= self:GetMaxJP() then 
				a = RED_NEON_COLOR
			elseif qtyJP > (self:GetMaxJP() - self.db.global.progress_orange) then 
				a = ORANGE_NEON_COLOR
			elseif qtyJP > (self:GetMaxJP() - self.db.global.progress_yellow) then 
				a = YELLOW_NEON_COLOR
			else
				a = GREEN_NEON_COLOR
			end
			jpText = a .. jpText .. END_COLOR
		end
	end
	
	-- Determine the LFR text to display
	if displaytype == 1 or displaytype == 2 or displaytype == 5 or displaytype == 7 or displaytype >= 9 then
		lfrDone, lfrTotal = self:GetLFRTotals(data.dungeon)
		
		if lfrTotal <= 0 then
			lfrText = "N/A";
		else
			lfrText = lfrDone .. "/" .. lfrTotal
		end

		if self:GetLDBTextColor() then
			a = GREEN_NEON_COLOR
			lfrText = a .. lfrText .. END_COLOR
		end
	end
	
	-- Join it all together now
	finalText = vpcapText
	
	if vpText ~= "" then
		closeit = false
		if finalText ~= "" then 
			finalText = finalText .. " (" 
			closeit = true
		end
		finalText = finalText .. vpText
		if closeit then finalText = finalText .. ")" end
	end
	
	if jpText ~= "" then
		closeit = false
		if finalText ~= "" then 
			finalText = finalText .. " <" 
			closeit = true
		end
		finalText = finalText .. jpText
		if closeit then finalText = finalText .. ">" end
	end
	
	if lfrText ~= "" then
		closeit = false
		if finalText ~= "" then 
			finalText = finalText .. " [" 
			closeit = true
		end
		finalText = finalText .. lfrText
		if closeit then finalText = finalText .. "]" end
	end
	
--print("what should be displayed on the LDB Text button .. " .. finalText)
	return " " .. finalText
end

function GG:GetLFRTotals(dungeons)
	
	local cntDone, cnt = 0, 0
	
	for i,v in pairs(dungeons) do
		if v.isAvailable and v.isAvailable >= 1 then
			cnt = cnt + 1
		end
		if v.isComplete and v.isComplete >= 1 then
			cntDone = cntDone + 1
		end
	end

	return cntDone, cnt
	
end

function GG:GetToonFilter() 
	
	local data = GG_Storage[self.currentRealm][self.currentPlayer]
	
	return data.active
end

function GG:GetMaxVP()
	if self.maxVP > 0 then
		return self.maxVP;
	else
		_, _, _, _, self.periodMaxVP, self.maxVP = GetCurrencyInfo(VALOR_CURRENCY_ID);
		self.periodMaxVP = floor(self.periodMaxVP / 100);
		self.maxVP = floor(self.maxVP/100);
		return self.maxVP;
	end

end

function GG:GetMaxJP()
	if self.maxJP > 0 then
		return self.maxJP;
	else
		_, _, _, _, _, self.maxJP = GetCurrencyInfo(JUSTICE_CURRENCY_ID);
		self.maxJP = floor(self.maxJP/100);
		return self.maxJP;
	end

end

function GG:GetPeriodMaxVP()
	if self.periodMaxVP > 0 then
		return self.periodMaxVP;
	else
		_, _, _, _, self.periodMaxVP, self.maxVP = GetCurrencyInfo(VALOR_CURRENCY_ID);
		self.periodMaxVP = floor(self.periodMaxVP / 100);
		self.maxVP = floor(self.maxVP/100);
		
		return self.periodMaxVP;
	end

end

function GG:GetResetHour() 
	if self.db.global.reset_hour < 0 or self.db.global.reset_hour > 23 then
		self.db.global.reset_hour = GG_Defaults.global.reset_hour;
	end
	
	return self.db.global.reset_hour; 
	
end

function GG:GetResetDayOfWeek() 
	if self.db.global.reset_dayofweek < 1 or self.db.global.reset_dayofweek > 7 then
		self.db.global.reset_dayofweek = GG_Defaults.global.reset_dayofweek;
	end
	
	return self.db.global.reset_dayofweek; 

end


-- Advanced sets
function GG:SetBGColor(r, g, b, a)
	local t = self.db.global.Background.Color
	t.r, t.g, t.b, t.a = r, g, b, a
end

function GG:SetBGBorderColor(r, g, b, a)
	local x = self.db.global.Background.BorderColor
	x.r, x.g, x.b, x.a = r, g, b, a
end

function GG:SetMinimap(v) 
    self.db.global.minimap.hide = not v;
	
	if not v then
		self.LDBIcon:Hide(ADDON_NAME);
	else
		self.LDBIcon:Show(ADDON_NAME);
	end
end

function GG:SetResetHour(v) 
	if v < 0 or v > 23 then
		self.db.global.reset_hour = GG_Defaults.global.reset_hour;
	else
		self.db.global.reset_hour = v; 
	end
	
end

function GG:SetResetDayOfWeek(v) 
	if v < 1 or v > 7 then
		self.db.global.reset_dayofweek = GG_Defaults.global.reset_dayofweek
	else
		self.db.global.reset_dayofweek = v
	end 

end

function GG:SetDebug(v) 

	self.db.char.debug = v
	
	if v then 
		self:Print("Entering debug mode")
	else 
		self:Print("Leaving debug mode")
	end

end

function GG:SetLDB(v)
	self.db.global.ldbText = v
	
	if v then
		self.LDBObj.type = 'data source'
	else
		self.LDBObj.type = 'launcher'
		self.LDBObj.text = 'n/a'
	end
	
	self:UpdateLDB(true);
end

function GG:SetVPPaperdoll(v)
	self.db.global.cinfoVP = v
	
	if v then
		PAPERDOLL_STATINFO[L.core.PDVPText] = { updateFunc = function(...) GG:UpdateVPPaperDollFrame(...); end }
	else
		PAPERDOLL_STATINFO[L.core.PDVPText] = nil
	end
end

function GG:SetVPCapPaperdoll(v)
	self.db.global.cinfoVPCap = v
	
	if v then
		PAPERDOLL_STATINFO[L.core.PDVPCapText] = { updateFunc = function(...) GG:UpdateVPCapPaperDollFrame(...); end }
	else
		PAPERDOLL_STATINFO[L.core.PDVPCapText] = nil
	end
end

function GG:SetJPPaperdoll(v)
	self.db.global.cinfoJP = v
	
	if v then
		PAPERDOLL_STATINFO[L.core.PDJPText] = { updateFunc = function(...) GG:UpdateJPPaperDollFrame(...); end }
	else
		PAPERDOLL_STATINFO[L.core.PDJPText] = nil
	end
end

function GG:SetToonFilter(v)

	local data = GG_Storage[self.currentRealm][self.currentPlayer]
	
	data.active = v

end

-- Advanced Updates
function GG:UpdateLDB(force, auto)
	
	self:SaveData()
	
	if self:GetLDB() then
		local label = UnitName('player')
		
		-- Do we really need to update LDB?
		--if force or label ~= self.ldbLabel then
            
            local t = self:GetButtonText()
			if t == nil then t = 'n/a' end
			self:UpdateLDBText(label, t)

			--end
	else
		self.LDBObj.type = 'launcher'
		self.LDBObj.text = 'n/a'
	end
end

function GG:UpdateLDBText(label, txt)
	if not self:GetLDB() then return false end
    
    -- Add the label
    if self:GetLDBlabel() then
        txt = label .. ": " .. txt
    end
    
	self.LDBObj.text = txt
end

function GG:UpdateVPPaperDollFrame(statFrame, unit)
    
	local formatted
	
	self:SaveData()
	
	PaperDollFrame_SetLabelAndText(statFrame, L.core.PDVPText, GG_Storage[self.currentRealm][self.currentPlayer].qtyVP, false)
    statFrame.tooltip = HIGHLIGHT_FONT_COLOR_CODE..L.core.PDVPText..FONT_COLOR_CODE_CLOSE
    if self:GetMaxVP() > 0 then
	  formatted = L.core.PDVPTooltip .. "\n" .. format(L.core.PDVPTooltipMax, self:GetMaxVP())
	else
	  formatted = L.core.PDVPTooltip
    end
    statFrame.tooltip2 = formatted
	
    statFrame:Show()
end

function GG:UpdateVPCapPaperDollFrame(statFrame, unit)
    
	local formatted;
	
	if self:GetPeriodMaxVP() > 0 then
	  formatted = GG_Storage[self.currentRealm][self.currentPlayer].periodEarnVP .. "/" .. self:GetPeriodMaxVP();
	else
	  formatted = GG_Storage[self.currentRealm][self.currentPlayer].periodEarnVP
	end
	
    PaperDollFrame_SetLabelAndText(statFrame, L.core.PDVPCapText, formatted, false)
    statFrame.tooltip = HIGHLIGHT_FONT_COLOR_CODE..L.core.PDVPCapText..FONT_COLOR_CODE_CLOSE
    if self:GetPeriodMaxVP() > 0 then
	  formatted = L.core.PDVPCapTooltip .. "\n" .. format(L.core.PDVPCapTooltipMax, self:GetPeriodMaxVP())
	else
	  formatted = L.core.PDVPCapTooltip
    end
    statFrame.tooltip2 = formatted
    
    statFrame:Show()
end

function GG:UpdateJPPaperDollFrame(statFrame, unit)
    	
    PaperDollFrame_SetLabelAndText(statFrame, L.core.PDJPText, GG_Storage[self.currentRealm][self.currentPlayer].qtyJP, false);
    statFrame.tooltip = HIGHLIGHT_FONT_COLOR_CODE..L.core.PDJPText..FONT_COLOR_CODE_CLOSE;
    if self:GetMaxJP() > 0 then
	  formatted = L.core.PDJPTooltip .. "\n" .. format(L.core.PDJPTooltipMax, self:GetMaxJP());
	else
	  formatted = L.core.PDJPTooltip;
    end
    statFrame.tooltip2 = formatted;
	
    statFrame:Show();
end

-- Debug function
function GG:Debug(...)
	if self:GetDebug() then
		print('|cFFFF0000GearGrinder Debug:|r ', ...);
	end
end

-- Open the options window
function GG:ShowOptions()
    -- Not using this until a paramiter is added to open expanded
    -- InterfaceOptionsFrame_OpenToCategory(L.core.name);
    
	self:SaveData();
	
    for i,element in pairs(INTERFACEOPTIONS_ADDONCATEGORIES) do 
        if element.name == L.core.name then
            if not InterfaceOptionsFrame:IsShown() then
                InterfaceOptionsFrame_Show();
                InterfaceOptionsFrameTab2:Click();
            end

            for i,button in pairs(InterfaceOptionsFrameAddOns.buttons) do
                if button.element == element then
                    InterfaceOptionsListButton_OnClick(button);
                    
                    -- Expand
                    if element.hasChildren and element.collapsed then
                        OptionsListButtonToggle_OnClick(button.toggle);
                    end
                
                -- Hide anything else that is open
                elseif button.element and type(button.element.collapsed) == 'boolean' and not button.element.collapsed then
                    OptionsListButtonToggle_OnClick(button.toggle);
                end
            end
        end
    end
end

